/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

This code utilises the work of the FreeType Project under the FreeType
License.  For more information please refer to the FreeType home page at
www.freetype.org.

******************************************************************************

See "freetype/include/freetype/config/ftoption.h" to alter FreeType options.

Modifications
-------------
Only .ttf fonts are supported through this implementation of the FreeType
library.

*****************************************************************************/

#define LOCAL_FONT

#include <ft2build.h>
#include FT_FREETYPE_H
#undef NULL
#define NULL 0

#include <pandora/data/config.h>
#include <pandora/files/file.h>
#include <pandora/system/all.h>
#include <pandora/graphics/font.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/picture.h>
#include <pandora/modules/font.h>
#include <pandora/modules/bitmap.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>

#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDOpen(OBJECTPTR);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, CMDOpen, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FontModule, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1998-2003.  All rights reserved.",
   "September 2003",
   "Font"
};

#define CHAR_TAB     (0x09)
#define CHAR_ENTER   (10)
#define CHAR_SPACE   '.'     /* Character to use for determining the size of a space */
#define SCREEN_DPI   100     /* 75 and 100 DPI seem to give the best measured results */
#define FT_DOWNSIZE  6

/******************************************************************************
** Win32 font structures
*/

struct FontList {
   LONG Offset, Size, Point;
};

struct winmz_header_fields {
   UWORD magic;
   UBYTE data[29 * 2];
   ULONG lfanew;
};

struct winne_header_fields {
   UWORD magic;
   UBYTE data[34];
   UWORD resource_tab_offset;
   UWORD rname_tab_offset;
};

struct winfnt_header_fields {
   UWORD version;
   ULONG file_size;
   UBYTE copyright[60];
   UWORD file_type;
   UWORD nominal_point_size;     /* Point size */
   UWORD vertical_resolution;
   UWORD horizontal_resolution;
   UWORD ascent;                 /* The amount of pixels above the base-line */
   UWORD internal_leading;       /* top leading pixels */
   UWORD external_leading;       /* gutter */
   BYTE  italic;                 /* TRUE if font is italic */
   BYTE  underline;              /* TRUE if font is underlined */
   BYTE  strike_out;             /* TRUE if font is striked-out */
   UWORD weight;                 /* Indicates font boldness */
   BYTE  charset;
   UWORD pixel_width;
   UWORD pixel_height;
   BYTE  pitch_and_family;
   UWORD avg_width;
   UWORD max_width;
   UBYTE first_char;
   UBYTE last_char;
   UBYTE default_char;
   UBYTE break_char;
   UWORD bytes_per_row;
   ULONG device_offset;
   ULONG face_name_offset;
   ULONG bits_pointer;
   ULONG bits_offset;
   BYTE  reserved;
   ULONG flags;
   UWORD A_space;
   UWORD B_space;
   UWORD C_space;
   UWORD color_table_offset;
   BYTE  reservedend[4];
} __attribute__((packed));

#define ID_WINMZ  0x5A4D
#define ID_WINNE  0x454E

/******************************************************************************
** Structure definition for cached bitmap fonts.
*/

struct BitmapCache {
   APTR Data;
   UBYTE *Outline;
   struct winfnt_header_fields Header;
   struct FontCharacter Chars[256];
   BYTE Location[200];
   WORD OpenCount;
   LONG StyleFlags;
};

static struct BitmapCache *glBitmapCache = 0;
static WORD glCacheSize;

/****************************************************************************/

struct KernelBase  *KernelBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR FontClass = 0, FontServerClass = 0;
static OBJECTPTR modStrings = 0, modFileSystem = 0, modBitmap = 0;
static OBJECTID glSystemFontsID = 0;
static FT_Library glFTLibrary = NULL;
static UBYTE *glAlphaLookup;

static void DoWordWrap(struct Font *);
static void CalcLineCount(struct Font *);
static ERROR AddFontClass(void);
static ERROR AddFontServerClass(void);
static LONG GetUTF8(struct Font *Self, STRING str, LONG *Unicode);
static LONG GetKerning(FT_Face Face, LONG Glyph, LONG PrevGlyph);

/*****************************************************************************
** Command:  Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR server;

   KernelBase = argKernelBase;

   /*** File System module ***/

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_NewObject));

   /*** Strings module ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_NewObject));

   /*** Bitmap module (we don't use any functions, we just want to create the alpha blending table) ***/

   if (CreateObject(ID_MODULE, NULL, &modBitmap, NULL,
      FID_Name|TSTRING,   "bitmap",
      FID_Version|TFLOAT, MODVERSION_BITMAP,
      TAGEND) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_NewObject));
   }

   /*** Initialise the freetype library ***/

   if (FT_Init_FreeType(&glFTLibrary)) {
      DPrintF("!Module:","[Font] Failed to initialise the FreeType font library.");
      return(ERR_Failed);
   }

   /*** Create the font server class ***/

   AddFontServerClass();

   /* Create the font server (required for looking up font faces and so
   ** forth during font initialisations).
   */

   if (NewNamedObject(ID_FONTSERVER, NF_UNTRACKED|NF_PUBLIC|NF_UNIQUE, &server, &glSystemFontsID, "SystemFonts") IS ERR_Okay) {
      Action(AC_Init, server, NULL);
      ReleaseObject(server);
   }

   if (AccessMemory(RPM_AlphaBlend, MEM_READWRITE|MEM_NOBLOCKING, &glAlphaLookup) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AccessMemory));
   }

   return(AddFontClass());
}

/*****************************************************************************
** Command: Open()
*/

static ERROR CMDOpen(OBJECTPTR Module)
{
   return(ERR_Okay);
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (glAlphaLookup) { ReleaseMemory(glAlphaLookup); glAlphaLookup = NULL; }

   /*** Close the FreeType library ***/

   FT_Done_FreeType(glFTLibrary);

   /*** Free alloated class and modules ***/

   if (FontClass)       { Action(AC_Free, FontClass, NULL); FontClass = NULL; }
   if (FontServerClass) { Action(AC_Free, FontServerClass, NULL); FontServerClass = NULL; }
   if (modFileSystem)   { Action(AC_Free, modFileSystem, NULL); modFileSystem = NULL; }
   if (modStrings)      { Action(AC_Free, modStrings, NULL); modStrings = NULL; }
   if (modBitmap)       { Action(AC_Free, modBitmap, NULL); modBitmap = NULL; }

   /* NB: Cached font files are not removed during expunge, because the task's
   **     shutdown procedure will have automatically destroy any cached fonts
   **     our CMDExpunge() routine is called.
   */

   if (glBitmapCache) FreeMemory(glBitmapCache);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: GetUTF8()
** Short:    Return the first unicode value from a given string address.
*/

inline static LONG GetUTF8(struct Font *Self, STRING str, LONG *Unicode)
{
   LONG i, len, code;

   if (*str < 128) {
      if (Unicode) *Unicode = *str;
      return(1);
   }
   else if ((*str & 0xe0) IS 0xc0) {
      len  = 2;
      code = *str & 0x1f;
   }
   else if ((*str & 0xf0) IS 0xe0) {
      len  = 3;
      code = *str & 0x0f;
   }
   else if ((*str & 0xf8) IS 0xf0) {
      len  = 4;
      code = *str & 0x07;
   }
   else if ((*str & 0xfc) IS 0xf8) {
      len  = 5;
      code = *str & 0x03;
   }
   else if ((*str & 0xfc) IS 0xfc) {
      len  = 6;
      code = *str & 0x01;
   }
   else {
      /*** Unprintable character ***/
      if (Unicode) *Unicode = 0;
      return(1);
   }

   for (i=1; i < len; ++i) {
      if ((str[i] & 0xc0) != 0x80) code = -1;
      code <<= 6;
      code |= str[i] & 0x3f;
   }

   if (code IS -1) {
      if (Unicode) *Unicode = 0;
      return(1);
   }
   else {
      if (Unicode) *Unicode = code;
      return(len);
   }
}

/****************************************************************************/

#include "obj_font.c"
#include "obj_fontserver.c"
